unit UDisassembler;

{$mode delphi}

interface

uses
  Classes, SysUtils, UMemory, fgl, UCpu, UInstructionReader, Math, UOpcode;

type
  TDisassemblerLine = class
    Address: Integer;
    Opcode: string;
    Instruction: string;
  end;

  { TDisassembler }

  TDisassembler = class(TInstructionReader)
  private
    OpcodeDefs: TOpcodeDefs;
  public
    Output: TFPGObjectList<TDisassemblerLine>;
    procedure Process;
    constructor Create;
    destructor Destroy; override;
  end;


implementation

{ TDisassembler }

procedure TDisassembler.Process;
var
  Opcode: Byte;
  Line: TDisassemblerLine;
  MemorySize: Integer;
  Data: QWord;
  Address: TAddress;
  AddressRel: TAddressSigned;
  Reg: TRegIndex;
  OpcodeDef: TOpcodeDef;
begin
  Init;
  Output.Clear;

  MemorySize := MemSize(Cpu.Memory);
  while IP < MemorySize do begin
    Prefix := False;
    Opcode := Read8;
    if Opcode <= Integer(High(TOpcode)) then begin
      OpcodeDef := OpcodeDefs.SearchByOpcode(TOpcode(Opcode));
      Prefix := OpcodeDef.Prefix;
      case TOpcode(Opcode) of
        opDataPrefix8: DataSize := bw8;
        opDataPrefix16: DataSize := bw16;
        opDataPrefix32: DataSize := bw32;
        opDataPrefix64: DataSize := bw64;
        opAddrPrefix8: AddrSize := bw8;
        opAddrPrefix16: AddrSize := bw16;
        opAddrPrefix32: AddrSize := bw32;
        opAddrPrefix64: AddrSize := bw64;
      end;
      Line := TDisassemblerLine.Create;
      Line.Address := IP - 1;
      Line.Opcode := IntToHex(Opcode, 2);
      Line.Instruction := OpcodeDef.Name;
      case OpcodeDef.Param1 of
        prReg: begin
          Reg := Read8;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Reg, 2);
          Line.Instruction := Line.Instruction + ' R' + IntToStr(Reg);
        end;
        prData: begin
          Data := ReadData;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Data, BitWidthBytes[DataSize] * 2);
          Line.Instruction := Line.Instruction + ' ' + IntToHexEx(Data, -1, '$');
        end;
        prAddr: begin
          Address := ReadAddress;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Address, BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ' ' + IntToHexEx(Address, -1, '$');
        end;
        prAddrRel: begin
          AddressRel := ReadAddressSigned;
          Line.Opcode := Line.Opcode + ' ' + IntToHexEx(QWord(AddressRel), BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ' ' + IntToHexEx(AddressRel, -1, '$');
        end;
      end;
      case OpcodeDef.Param2 of
        prReg: begin
          Reg := Read8;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Reg, 2);
          Line.Instruction := Line.Instruction + ', R' + IntToStr(Reg);
        end;
        prData: begin
          Data := ReadData;
          Line.Opcode := Line.Opcode + ' ' + IntToHexEx(Data, BitWidthBytes[DataSize] * 2);
          Line.Instruction := Line.Instruction + ', ' + IntToHexEx(Data, -1, '$');
        end;
        prAddr: begin
          Address := ReadAddress;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Address, BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ', ' + IntToHexEx(Address, -1, '$');
        end;
        prAddrRel: begin
          AddressRel := ReadAddressSigned;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(QWord(AddressRel), BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ', ' + IntToHexEx(AddressRel, -1, '$');
        end;
      end;
      case OpcodeDef.Param3 of
        prReg: begin
          Reg := Read8;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Reg, 2);
          Line.Instruction := Line.Instruction + ', R' + IntToStr(Reg);
        end;
        prData: begin
          Data := ReadData;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Data, BitWidthBytes[DataSize] * 2);
          Line.Instruction := Line.Instruction + ', ' + IntToHexEx(Data, -1, '$');
        end;
        prAddr: begin
          Address := ReadAddress;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(Address, BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ', ' + IntToHexEx(Address, -1, '$');
        end;
        prAddrRel: begin
          AddressRel := ReadAddressSigned;
          Line.Opcode := Line.Opcode + ' ' + IntToHex(QWord(AddressRel), BitWidthBytes[AddrSize] * 2);
          Line.Instruction := Line.Instruction + ', ' + IntToHexEx(AddressRel, -1, '$');
        end;
      end;
      Output.Add(Line);
      if not Prefix then begin
        DataSize := DataSizeBase;
        AddrSize := AddrSizeBase;
      end;
    end else begin
      {Line := TDisassemblerLine.Create;
      Line.Address := IP;
      Line.Opcode := IntToHex(Opcode, 2);
      Line.Instruction := '';
      Output.Add(Line);
      }
    end;
  end;
end;

constructor TDisassembler.Create;
begin
  OpcodeDefs := TOpcodeDefs.Create;
  Output := TFPGObjectList<TDisassemblerLine>.Create;
end;

destructor TDisassembler.Destroy;
begin
  FreeAndNil(Output);
  FreeAndNil(OpcodeDefs);
  inherited Destroy;
end;


end.

