unit UInstructionWriter;

{$mode delphi}{$H+}

interface

uses
  Classes, SysUtils, UCpu;

type

  { TInstructionWriter }

  TInstructionWriter = class
  private
    procedure PrefixBegin;
    procedure PrefixEnd;
  public
    Cpu: TCpu;
    IP: Integer;
    DataSize: TBitWidth;
    DataSizeBase: TBitWidth;
    AddrSize: TBitWidth;
    AddrSizeBase: TBitWidth;
    Prefix: Boolean;
    procedure Init;
    function GetRelativeAddr(BitWidth: TBitWidth; BaseIP, TargetAddr: QWord): Int64;
    procedure Write8(Value: Byte);
    procedure Write16(Value: Word);
    procedure Write32(Value: DWord);
    procedure Write64(Value: QWord);
    procedure WriteAddress(Value: TAddress);
    procedure WriteAddressSigned(Value: TAddressSigned); inline;
    procedure WriteData(Value: QWord);
    procedure WriteString(Text: string);
    procedure Nop;
    procedure Halt;
    procedure Loadi(Reg: TRegIndex; Value: QWord);
    procedure Increment(Reg: TRegIndex);
    procedure Decrement(Reg: TRegIndex);
    procedure Jump(Addr: QWord);
    procedure JumpNotZero(Addr: QWord);
    procedure JumpZero(Addr: QWord);
    procedure JumpRel(Addr: QWord);
    procedure JumpRelNotZero(Addr: QWord);
    procedure JumpRelZero(Addr: QWord);
    procedure Push(Reg: TRegIndex);
    procedure Pop(Reg: TRegIndex);
    procedure Call(Addr: QWord);
    procedure Return;
    procedure Convert(Reg: TRegIndex);
    procedure DataPrefix8;
    procedure DataPrefix16;
    procedure DataPrefix32;
    procedure DataPrefix64;
    procedure AddrPrefix8;
    procedure AddrPrefix16;
    procedure AddrPrefix32;
    procedure AddrPrefix64;
    procedure StoreMem(RegAddr, RegSrc: TRegIndex);
    procedure LoadMem(RegDst, RegAddr: TRegIndex);
    procedure Output(Port: TAddress; Reg: TRegIndex);
    procedure Input(Reg: TRegIndex; Port: TAddress);
    procedure Test(Reg: TRegIndex);
  end;


implementation

{ TInstructionWriter }

procedure TInstructionWriter.Nop;
begin
  PrefixBegin;
  Write8(Byte(opNop));
  PrefixEnd;
end;

procedure TInstructionWriter.Halt;
begin
  PrefixBegin;
  Write8(Byte(opHalt));
  PrefixEnd;
end;

procedure TInstructionWriter.Loadi(Reg: TRegIndex; Value: QWord);
begin
  PrefixBegin;
  Write8(Byte(opLoadi));
  Write8(Reg);
  WriteData(Value);
  PrefixEnd;
end;

procedure TInstructionWriter.Increment(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opInc));
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.Decrement(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opDec));
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.Jump(Addr: QWord);
begin
  PrefixBegin;
  Write8(Byte(opJump));
  WriteAddress(Addr);
  PrefixEnd;
end;

procedure TInstructionWriter.JumpNotZero(Addr: QWord);
begin
  PrefixBegin;
  Write8(Byte(opJumpNotZero));
  WriteAddress(Addr);
  PrefixEnd;
end;

procedure TInstructionWriter.JumpZero(Addr: QWord);
begin
  PrefixBegin;
  Write8(Byte(opJumpZero));
  WriteAddress(Addr);
  PrefixEnd;
end;

function TInstructionWriter.GetRelativeAddr(BitWidth: TBitWidth; BaseIP, TargetAddr: QWord): Int64;
begin
  Result := Int64(TargetAddr) - (BaseIP + 1 + BitWidthBytes[BitWidth]);
end;

procedure TInstructionWriter.JumpRel(Addr: QWord);
var
  RelAddr: Int64;
begin
  PrefixBegin;
  RelAddr := GetRelativeAddr(AddrSize, IP, Addr);
  Write8(Byte(opJumpRel));
  WriteAddressSigned(RelAddr);
  PrefixEnd;
end;

procedure TInstructionWriter.JumpRelNotZero(Addr: QWord);
var
  RelAddr: Int64;
begin
  PrefixBegin;
  RelAddr := GetRelativeAddr(AddrSize, IP, Addr);
  Write8(Byte(opJumpRelNotZero));
  WriteAddressSigned(RelAddr);
  PrefixEnd;
end;

procedure TInstructionWriter.JumpRelZero(Addr: QWord);
var
  RelAddr: Int64;
begin
  PrefixBegin;
  RelAddr := GetRelativeAddr(AddrSize, IP, Addr);
  Write8(Byte(opJumpRelZero));
  WriteAddressSigned(RelAddr);
  PrefixEnd;
end;

procedure TInstructionWriter.Push(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opPush));
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.Pop(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opPop));
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.Call(Addr: QWord);
begin
  PrefixBegin;
  Write8(Byte(opCall));
  WriteAddress(Addr);
  PrefixEnd;
end;

procedure TInstructionWriter.Return;
begin
  PrefixBegin;
  Write8(Byte(opRet));
  PrefixEnd;
end;

procedure TInstructionWriter.Convert(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opConvert));
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.DataPrefix8;
begin
  Prefix := True;
  DataSize := bw8;
  Write8(Byte(opDataPrefix8));
end;

procedure TInstructionWriter.DataPrefix16;
begin
  Prefix := True;
  DataSize := bw16;
  Write8(Byte(opDataPrefix16));
end;

procedure TInstructionWriter.DataPrefix32;
begin
  Prefix := True;
  DataSize := bw32;
  Write8(Byte(opDataPrefix32));
end;

procedure TInstructionWriter.DataPrefix64;
begin
  Prefix := True;
  DataSize := bw64;
  Write8(Byte(opDataPrefix64));
end;

procedure TInstructionWriter.AddrPrefix8;
begin
  Prefix := True;
  AddrSize := bw8;
  Write8(Byte(opAddrPrefix8));
end;

procedure TInstructionWriter.AddrPrefix16;
begin
  Prefix := True;
  AddrSize := bw16;
  Write8(Byte(opAddrPrefix16));
end;

procedure TInstructionWriter.AddrPrefix32;
begin
  Prefix := True;
  AddrSize := bw32;
  Write8(Byte(opAddrPrefix32));
end;

procedure TInstructionWriter.AddrPrefix64;
begin
  Prefix := True;
  AddrSize := bw64;
  Write8(Byte(opAddrPrefix64));
end;

procedure TInstructionWriter.StoreMem(RegAddr, RegSrc: TRegIndex);
begin
  Write8(Byte(opStoreMem));
  Write8(RegAddr);
  Write8(RegSrc);
end;

procedure TInstructionWriter.LoadMem(RegDst, RegAddr: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opLoadMem));
  Write8(RegDst);
  Write8(RegAddr);
  PrefixEnd;
end;

procedure TInstructionWriter.Output(Port: TAddress; Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opOut));
  WriteAddress(Port);
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.Input(Reg: TRegIndex; Port: TAddress);
begin
  PrefixBegin;
  Write8(Byte(opIn));
  Write8(Reg);
  WriteAddress(Port);
  PrefixEnd;
end;

procedure TInstructionWriter.Test(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opTest));
  Write8(Reg);
  Prefixend;
end;

procedure TInstructionWriter.WriteString(Text: string);
var
  I: Integer;
begin
  for I := 1 to Length(Text) do
    Write8(Ord(Text[I]));
end;

procedure TInstructionWriter.PrefixBegin;
begin
  Prefix := False;
end;

procedure TInstructionWriter.PrefixEnd;
begin
  if not Prefix then begin
    DataSize := DataSizeBase;
    AddrSize := AddrSizeBase;
  end;
end;

procedure TInstructionWriter.Init;
begin
  DataSizeBase := Cpu.DataSizeBase;
  DataSize := DataSizeBase;
  AddrSizeBase := Cpu.AddrSizeBase;
  AddrSize := AddrSizeBase;
  IP := 0;
  Prefix := False;
end;

procedure TInstructionWriter.Write8(Value: Byte);
begin
  PByte(Cpu.Memory + IP)^ := Value;
  Inc(IP, SizeOf(Byte));
end;

procedure TInstructionWriter.Write16(Value: Word);
begin
  PWord(Cpu.Memory + IP)^ := Value;
  Inc(IP, SizeOf(Word));
end;

procedure TInstructionWriter.Write32(Value: DWord);
begin
  PDWord(Cpu.Memory + IP)^ := Value;
  Inc(IP, SizeOf(DWord));
end;

procedure TInstructionWriter.Write64(Value: QWord);
begin
  PQWord(Cpu.Memory + IP)^ := Value;
  Inc(IP, SizeOf(QWord));
end;

procedure TInstructionWriter.WriteAddress(Value: TAddress);
begin
  case AddrSize of
    bw8: Write8(Value and $ff);
    bw16: Write16(Value and $ffff);
    bw32: Write32(Value and $ffffffff);
    bw64: Write64(Value);
  end;
end;

procedure TInstructionWriter.WriteAddressSigned(Value: TAddressSigned);
begin
  case AddrSize of
    bw8: Write8(Byte(Value and $ff));
    bw16: Write16(Word(Value and $ffff));
    bw32: Write32(DWord(Value and $ffffffff));
    bw64: Write64(QWord(Value));
  end;
end;

procedure TInstructionWriter.WriteData(Value: QWord);
begin
  case DataSize of
    bw8: Write8(Value and $ff);
    bw16: Write16(Value and $ffff);
    bw32: Write32(Value and $ffffffff);
    bw64: Write64(Value);
  end;
end;

end.

